{ *********************************************************************** }
{                                                                         }
{ Delphi Runtime Library                                                  }
{ Huffman Compress/Decompress                                             }
{                                                                         }
{ Copyright (c) 2002 Poirot Programming Investigator                      }
{                                                                         }
{ *********************************************************************** }

unit huffman;

interface
type
  TpavFileName = string; //My File Name Type
  TpavAssignature = array[0..3] of Char; //type for Assignature

  PpavUzel = ^TpavUzel; //Pointor to TpavUzel Type
  TpavUzel = record //Element of the Huffmen Tree
    chSimbol: char;
    carCount: Cardinal;
    pLeft, pRight: PpavUzel;
  end;

  TpavPACKOptions = type Word;
  PpavPACKOptions = ^TpavPACKOptions;
  {bit Options Mask
    1 - bool256CharsIn (0-False);
    2-4 - bAddBits (7-maximum);
    4-16 - Reserved
  }
  PpavPACKOptionsRec = ^TpavPACKOptionsRec;
  TpavPACKOptionsRec = record
    bool256CharsIn: Boolean;
    bAddBits: Byte;
  end;

  TpavPACKHeader = record //PACK Header format
    sAssign: TpavAssignature;
    bOptions: TpavPACKOptions;
    bcrcSource: Cardinal;
    bCharCount: byte;
  end;

  TpavCharHash = array[char] of Cardinal; //Char's Ratio Array
  TpavPUzelArray = array of PpavUzel; //Tree-Base Array
  TpavCodesStr = string; //Byte-Str bit mask
  TpavCodes = array[char] of TpavCodesStr; //Pack Codes Table

const
  TREE_GOTO_LEFT = #0;
  TREE_GOTO_RIGHT = #1;
  PACK_ASSIGNATURE: TpavAssignature = 'CDD!'; //PACKED File Assignature

  //All-Purpose Proc/Func
//Str(Bin Mask) to Byte
procedure Common_BinToByte(var ABinMask: TpavCodesStr; out AOutByte: Byte; AStartBit: Byte = 1; AEndBit: byte = 8);
//Str(Bin Mask) to Byte
procedure Common_ByteToBin(AInByte: Byte; out AOutBinMask: TpavCodesStr);
//CRC32 Table Method Calculator from DLL
function Common_CRC32(AFileNAme: PChar; AStartValue: Cardinal; AReverce: Boolean): Cardinal; stdcall;
//Get Temp Dir Path
function Common_GetTempPath(nBufferLength: LongWord; lpBuffer: PChar): LongWord; stdcall;

//for Compression
//Analising of non-Compression File
procedure File_FillCharHash(var ASourceF: file; var ACharHash: TpavCharHash);
//Calculate Count of AddBits
procedure File_CalcAddBits(out AAddBits: Byte; var ACharHash: TpavCharHash; var ACodes: TpavCodes);
//Fill Options From Record
procedure File_EncodeOptions(AOptionsR: TpavPACKOptionsRec; out AOptions: TpavPACKOptions);
//Write to Target File Header (Assignature, CRC32, Table);
procedure File_WriteHeader(var ATargetF: file; var AHeader: TpavPACKHeader; var ACharHash: TpavCharHash);
//Compression to ATagret from ASource use ACodes
procedure File_Compress(var ATargetF, ASourceF: file; AAddBits: Byte; var ACodes: TpavCodes);

//for Decompression
//Fill Options From 16 bits
procedure File_DecodeOptions(AOptions: TpavPACKOptions; AOptionsR: PpavPACKOptionsRec);
//Read Packed File Header
procedure File_ReadHeader(var ASourceF: file; out AHeader: TpavPACKHeader; out ACharHash: TpavCharHash);
//Decode One Simbol
function File_DecodeOneSimbol(AHead: PpavUzel; var ABitMask: TpavCodesStr): Char;
//DeCompression to ATagret from ASource use ACodes
procedure File_DeCompress(var ATargetF, ASourceF: file; AHead: PpavUzel; AAddBits: Byte);

//For Tree
//Fill non-zero chars of array
procedure Tree_FillBase(var ACharHash: TpavCharHash; var ABase: TpavPUzelArray; out ACharCount: Byte);
//Getting 2 minimus elemnts indexses
procedure Tree_TwoMin(ABase: TpavPUzelArray; ALen: Integer; out AMinL, AMinR: Integer);
//Build Tree With ABase Array
function Tree_Build(var ABase: TpavPUzelArray; ALen: Integer): PpavUzel;
//Get Ecuals for Chars in Binar Codes
procedure Tree_GetCodes(AHead: PpavUzel; var ACodes: TpavCodes; ACode: TpavCodesStr = '');
//deleteTree from Head
procedure Tree_Delete(AHead: PpavUzel);

implementation

//////////////////////////////////////////////////////////
//All-Purpose Proc/Func
//////////////////////////////////////////////////////////

//Str(Bin Mask) to Byte

procedure Common_BinToByte;
var
  I, Sum: Byte;
begin
  Sum := 0;
  for I := AStartBit to AEndBit - 1 do
    Sum := (Sum + Byte(ABinMask[I])) shl 1;
  Sum := (Sum + Byte(ABinMask[AEndBit]));
  AOutByte := Sum;
end;

//Str(Bin Mask) to Byte

procedure Common_ByteToBin;
var I: Byte;
begin
  AOutBinMask := '00000000';
  for I := 8 downto 1 do begin
    if AInByte mod 2 = 0 then AOutBinMask[I] := TREE_GOTO_LEFT
    else AOutBinMask[I] := TREE_GOTO_RIGHT;
    AInByte := AInByte div 2;
  end;
end;

//CRC32 Table Method Calculator from DLL
function Common_CRC32; external 'md_crc32.dll' name 'CRC32_Calc';

//Get Temp Dir Path
function Common_GetTempPath; external 'kernel32.dll' name 'GetTempPathA';
//////////////////////////////////////////////////////////
//for Compression
//////////////////////////////////////////////////////////

//Analising of non-Compression File

procedure File_FillCharHash;
var
  Buf: Char;
  NumRead: Integer;
begin
  for Buf := #0 to #255 do ACharHash[Buf] := 0;
  repeat
    BlockRead(ASourceF, buf, SizeOf(Buf), NumRead);
    if NumRead <> 0 then
      ACharHash[Buf] := ACharHash[Buf] + 1;
  until NumRead = 0;
end;

//Calculate Count of AddBits

procedure File_CalcAddBits;
var
  I: Char;
  Sum: Cardinal;
begin
  Sum := 0;
  for I := #0 to #255 do
    Sum := Sum + ACharHash[I] * Length(ACodes[I]);
  AAddBits := 8 - (Sum mod 8);
  if AAddBits = 8 then AAddBits := 0;
end;

//Fill Options From Record

procedure File_EncodeOptions;
var
  I, Tmp: Byte;
  buf: TpavPACKOptions;
begin
  buf := $0000; {32 bits}
  //add 256CharsIn
  buf := (buf + Ord(AOptionsR.bool256CharsIn)) shl 1;
  //add AddBits Count
  Tmp := AOptionsR.bAddBits;
  for I := 2 downto 1 do begin
    buf := (buf + Tmp div (I * 2)) shl 1;
    Tmp := Tmp mod (I * 2);
  end;
  buf := buf + Tmp;
  //End shl to 256CharsIn mast be a first BIT
  buf := buf shl 12;
  //Result of Calculations Out
  AOptions := buf;
end;

//Write to Target File Header (Assignature, CRC32, Table);

procedure File_WriteHeader;
var
  bufCar: Cardinal;
  I, bufChar: Char;
  Tmp: Boolean; //Is Hash Have 256 Items
begin
  Tmp := AHeader.bOptions shr 15 = 1;
  with AHeader do begin
    BlockWrite(ATargetF, sAssign, SizeOf(sAssign)); //Assignatur
    BlockWrite(ATargetF, bOptions, SizeOf(bOptions)); //Options
    BlockWrite(ATargetF, bcrcSource, SizeOf(bcrcSource)); //CRC32 fo Source file
    if not Tmp then
      BlockWrite(ATargetF, bCharCount, SizeOf(bCharCount)); //char Hash Table Size
  end;
  //Char Hash Table
  for I := #0 to #255 do
    if ACharHash[I] <> 0 then begin
      bufChar := I;
      bufCar := ACharHash[I];
      if not Tmp then
        BlockWrite(ATargetF, bufChar, SizeOf(bufChar));
      BlockWrite(ATargetF, bufCar, SizeOf(bufCar) - 1);
    end;
end;

//Compression to ATagret from ASource use ACodes

procedure File_Compress;
var
  tmp: TpavCodesStr;
  NumRead: integer;
  bufr: char;
  bufw: byte;
begin
  Tmp := '';
  repeat
    BlockRead(ASourceF, bufr, SizeOf(bufr), NumRead);
    if NumRead <> 0 then begin
      Tmp := Tmp + ACodes[bufr];
      if Length(Tmp) >= 8 then begin
        Common_BinTobyte(Tmp, bufw);
        Delete(Tmp, 1, 8);
        BlockWrite(ATargetF, bufw, SizeOf(bufw));
      end;
    end;
  until NumRead = 0;
  if AAddBits <> 0 then begin
    for bufw := 1 to AAddBits do Tmp := Tmp + TREE_GOTO_LEFT;
    Common_BinToByte(Tmp, bufw);
    BlockWrite(ATargetF, bufw, SizeOf(bufw));
  end;
end;

//////////////////////////////////////////////////////////
//for Decompression
//////////////////////////////////////////////////////////

//Fill Options From 16 bits

procedure File_DecodeOptions;
begin
  with AOptionsR^ do begin
    bool256CharsIn := AOptions > $8000;
    bAddBits := (AOptions shl 1) shr 13;
  end;
end;

//Read Packed File Header

procedure File_ReadHeader;
var
  I, bufChar: Byte;
  J: Char;
  bufCar: Cardinal;
  Tmp: Boolean; //Is Hash Have 256 Items
begin
  with AHeader do begin
    BlockRead(ASourceF, sAssign, SizeOf(sAssign));
    BlockRead(ASourceF, bOptions, SizeOf(bOptions));
    Tmp := bOptions shr 15 = 1;
    BlockRead(ASourceF, bcrcSource, SizeOf(bcrcSource));
    if not Tmp then
      BlockRead(ASourceF, bCharCount, SizeOf(bCharCount))
    else bCharCount := 255;
  end;

  for J := #0 to #255 do ACharHash[J] := 0;
  for I := 0 to AHeader.bCharCount do begin
    if not Tmp then
      BlockRead(ASourceF, bufChar, SizeOf(bufChar))
    else bufChar:=I;
    BlockRead(ASourceF, bufCar, SizeOf(bufCar) - 1);
    ACharHash[Char(bufChar)] := bufCar;
  end;
end;

//Decode One Simbol

function File_DecodeOneSimbol;
var
  Cur: PpavUzel;
begin
  Result := #0;
  if AHead = nil then Exit;
  Cur := AHead;
  while (Cur^.pLeft <> nil) and (Cur^.pRight <> nil) do begin
    if ABitMask[1] = TREE_GOTO_LEFT then Cur := Cur^.pLeft
    else Cur := Cur^.pRight;
    Delete(ABitMask, 1, 1);
  end;
  Result := Cur^.chSimbol;
end;

//DeCompression to ATagret from ASource use ACodes

procedure File_DeCompress;
var
  DecStr: string;
  OneByte: TpavCodesStr;
  NumRead: Integer;
  Buf: Byte;
  OutChar: Char;
begin
  DecStr := '';
  OneByte := '';
  repeat
    BlockRead(ASourceF, Buf, SizeOf(Buf), NumRead);
    if NumRead <> 0 then begin
      Common_ByteToBin(Buf, OneByte);
      DecStr := DecStr + OneByte;
      while Length(DecStr) >= 256 do begin
        OutChar := File_DecodeOneSimbol(AHead, DecStr);
        BlockWrite(ATargetF, OutChar, SizeOf(OutChar));
      end;
    end;
  until NumRead = 0;
  while Length(DecStr) > AAddBits do begin
    OutChar := File_DecodeOneSimbol(AHead, DecStr);
    BlockWrite(ATargetF, OutChar, SizeOf(OutChar));
  end;
end;
//////////////////////////////////////////////////////////
//for Tree
//////////////////////////////////////////////////////////

//Fill non-zero chars of array

procedure Tree_FillBase;
var
  I: Byte;
  tmp: PpavUzel;
  Chars: string;
begin
  Chars := '';
  for I := 0 to 255 do
    if ACharHash[Char(I)] <> 0 then
      Chars := Chars + Char(I);
  ACharCount := Length(Chars) - 1;
  SetLength(ABase, ACharCount + 1);
  for I := 0 to ACharCount do begin
    New(Tmp);
    with Tmp^ do begin
      chSimbol := Chars[I + 1];
      carCount := ACharHash[chSimbol];
      pLeft := nil;
      pRight := nil;
    end;
    ABase[I] := Tmp;
  end;
end;

//Getting 2 minimus elemnts indexses

procedure Tree_TwoMin;
var
  ml, mr, i: integer;
begin
  if ABase[0]^.carCount < ABase[1]^.carCount then begin
    ml := 0; mr := 1;
  end else begin
    ml := 1; mr := 0;
  end;
  for i := 2 to ALen - 1 do
    if (ABase[i]^.carCount < ABase[mr]^.carCount) then begin
      if (ABase[i]^.carCount <= ABase[ml]^.carCount) then begin
        mr := ml; ml := i;
      end else mr := i;
    end;
  AMinL := ml;
  AMinR := mr;
end;

//Build Tree With ABase Array

function Tree_Build;
var
  ml, mr, max_i, min_i, len: Integer;
  tmp: PpavUzel;
begin
  tmp := nil;
  len := ALen;
  while len > 1 do begin
    Tree_TwoMin(ABase, len, ml, mr);
    len := len - 1;
    New(tmp);
    if mr > ml then begin max_i := mr; min_i := ml; end
    else begin max_i := ml; min_i := mr; end;
    with tmp^ do begin
      chSimbol := ABase[max_i]^.chSimbol;
      carCount := ABase[max_i]^.carCount + ABase[min_i]^.carCount;
      pLeft := ABase[min_i];
      pRight := ABase[max_i];
    end;
    ABase[min_i] := tmp;
    ABase[max_i] := ABase[len];
    ABase[len] := nil;
    tmp := nil;
  end;
  Result := ABase[0];
end;

//Get Ecuals for Chars in Binar Codes

procedure Tree_GetCodes;
begin
  if AHead <> nil then begin
    Tree_GetCodes(AHead^.pLeft, ACodes, ACode + TREE_GOTO_LEFT);
    Tree_GetCodes(AHead^.pRight, ACodes, ACode + TREE_GOTO_RIGHT);
    if ACodes[AHead^.chSimbol] = '' then begin
      ACodes[AHead^.chSimbol] := ACode;
    end;

  end;
end;

//deleteTree from Head

procedure Tree_Delete;
begin
  if AHead <> nil then begin
    Tree_Delete(AHead^.pLeft);
    Tree_Delete(AHead^.pRight);
    Dispose(AHead);
  end
end;

end.

