library huffman_dll;

uses
  pav_PACK_DLL,
  huffman;

{$R *.res}

function PACK_DLL_Compress(ATarget, ASource: string): Byte; stdcall;
var
  CharHash: TpavCharHash; //Char Hash from Source File
  Base: TpavPUzelArray; //Base Array for Huffmen-Tree Building
  Codes: TpavCodes; //Codes for each simbol in source file
  Head: PpavUzel; //The Pointer to Root Element
  Header: TpavPACKHeader;
  I: Char; //Iterational variable
  ft, fs: file; //points to File Streams
  CharCount, AddBits: Byte;
  crcSource: Cardinal; //CRC32 code
  OptionsRec: TpavPACKOptionsRec; //Options Record
begin
  crcSource := Common_CRC32(PChar(ASource), $FFFFFFFF, true);
  AssignFile(fs, ASource);
  Reset(fs, 1);
  File_FillCharHash(fs, CharHash);
  CloseFile(fs);

  Tree_FillBase(CharHash, Base, CharCount);
  Head := Tree_Build(Base, Length(Base));
  for I := #0 to #255 do Codes[I] := '';
  Tree_GetCodes(Head, Codes, '');
  Tree_Delete(Head);

  File_CalcAddBits(AddBits, CharHash, Codes);
  Reset(fs, 1);
  AssignFile(ft, ATarget);
  Rewrite(ft, 1);

  {Compression of the file}
  //Set Options
  with OptionsRec do begin
    bool256CharsIn := (CharCount = 255);
    bAddBits := AddBits;
  end;
  //Encode Options To BitMask
  File_EncodeOptions(OptionsRec, Header.bOptions);
  //Fill Pack Header
  with Header do begin
    sAssign := PACK_ASSIGNATURE;
    bcrcSource := crcSource;
    bCharCount := CharCount;
  end;

  File_WriteHeader(ft, Header, CharHash);
  File_Compress(ft, fs, Addbits, Codes);

  CloseFile(ft);
  CloseFile(fs);
  Result := COMPRESS_OK;
end;

function PACK_DLL_Decompress(ATarget, ASource: string): Byte; stdcall;
var
  fs, ft: file;
  Base: TpavPUzelArray; //Base Array for Huffmen-Tree Building
  Head: PpavUzel; //The Pointer to Root Element
  Header: TpavPACKHeader;
  OptionsRec: TpavPACKOptionsRec; //Options Record
  CharCount: byte;
  CharHash: TpavCharHash;
  crcSource: Cardinal;
begin
  AssignFile(fs, ASource);
  Reset(fs, 1);
  if FileSize(fs) < SizeOf(Header) then begin
    Result := DECOMP_SIZE;
    CloseFile(fs);
    Exit;
  end;
  File_ReadHeader(fs, Header, CharHash);
  if Header.sAssign <> PACK_ASSIGNATURE then begin
    Result := DECOMP_ASSIGN;
    CloseFile(fs);
    Exit;
  end;
  Tree_FillBase(CharHash, Base, CharCount);
  File_DecodeOptions(Header.bOptions, @OptionsRec);
  Head := Tree_Build(Base, Length(Base));
  AssignFile(ft, ATarget);
  Rewrite(ft, 1);
  File_DeCompress(ft, fs, Head, OptionsRec.bAddBits);
  CloseFile(ft);
  CloseFile(fs);
  Tree_Delete(Head);
  crcSource := Common_CRC32(PChar(ATarget), $FFFFFFFF, true);
  if crcSource = Header.bcrcSource then Result := DECOMPRESS_OK
  else Result := DECOMP_CRC32;
end;

procedure PACK_DLL_ReadHeaderEx(AFileName: string; AHeaderEx: PpavFileHeaderEx); stdcall;
var
  fs: file;
  I: Char;
  FileSize: Cardinal;
  CharCount: Word;
  CharHash: TpavCharHash;
  Header: TpavPACKHeader;
  OptionsR: TpavPACKOptionsRec;
begin
  for I := #0 to #255 do CharHash[I] := 0;
  AssignFile(fs, AFileName);
  Reset(fs, 1);
  File_ReadHeader(fs, Header, CharHash);
  CloseFile(fs);
  File_DecodeOptions(Header.bOptions, @OptionsR);
  AHeaderEx^.bAddBits := OptionsR.bAddBits;
  CharCount := 0;
  FileSize := 0;
  for I := #0 to #255 do
    if CharHash[i] <> 0 then begin
      CharCount := CharCount + 1;
      FileSize := FileSize + CharHash[i];
    end;
  AHeaderEx^.bCharCount := CharCount - 1;
  AHeaderEx^.carFileSize := FileSize;
end;

exports
  PACK_DLL_Compress name PACK_DLL_Compress_Name,
  PACK_DLL_Decompress name PACK_DLL_Decompress_Name,
  PACK_DLL_ReadHeaderEx name PACK_DLL_ReadHeaderEx_Name;

begin
end.

